import io

from django.contrib import admin, messages
from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.decorators import permission_required
from django.core import management
from django.core.exceptions import PermissionDenied
from django.forms.models import modelform_factory
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.template import Context, Template
from django.urls import path, reverse
from django.utils.translation import gettext, gettext_lazy

from .models import Album, AlbumAccessPolicy, Photo, PhotoAccessPolicy


class SetAccessPolicyMixin:
    actions = ["set_access_policy", "unset_access_policy"]

    def set_access_policy(self, request, queryset):
        policy_model = {Album: AlbumAccessPolicy, Photo: PhotoAccessPolicy}[self.model]
        model_name = self.model._meta.model_name
        policy_model_name = policy_model._meta.model_name
        form_class = modelform_factory(
            policy_model,
            exclude=(model_name,),
            widgets={
                "users": FilteredSelectMultiple(gettext("Users"), False),
                "groups": FilteredSelectMultiple(gettext("Groups"), False),
            },
        )

        if request.POST.get("set_access_policy"):
            form = form_class(request.POST)
            if form.is_valid():
                queryset = queryset.select_related("access_policy")
                # Check permissions
                has_add_perm = request.user.has_perm(f"gallery.add_{policy_model_name}")
                has_change_perm = request.user.has_perm(
                    f"gallery.change_{policy_model_name}"
                )
                for obj in queryset:
                    try:
                        obj.access_policy
                    except policy_model.DoesNotExist:
                        if not has_add_perm:
                            raise PermissionDenied
                    else:
                        if not has_change_perm:
                            raise PermissionDenied
                # Apply changes
                created, changed = 0, 0
                for obj in queryset:
                    try:
                        ap = obj.access_policy
                        changed += 1
                    except policy_model.DoesNotExist:
                        ap = policy_model.objects.create(**{model_name: obj})
                        created += 1
                    form_class(request.POST, instance=ap).save()
                message = gettext(
                    "Successfully created %(created)d and "
                    "changed %(changed)d access policies."
                )
                message = message % {"created": created, "changed": changed}
                self.message_user(request, message)
                return HttpResponseRedirect(
                    reverse(f"admin:gallery_{model_name}_changelist")
                )
        else:
            form = form_class()
        context = {
            "action_checkbox_name": ACTION_CHECKBOX_NAME,
            "opts": self.model._meta,
            "form": form,
            "media": self.media + form.media,
            "photos": Photo.objects.filter(album__in=queryset),
            "queryset": queryset,
            "title": gettext("Set access policy"),
        }
        return render(request, "admin/gallery/set_access_policy.html", context)

    set_access_policy.short_description = gettext_lazy("Set access policy")

    def unset_access_policy(self, request, queryset):
        policy_model = {Album: AlbumAccessPolicy, Photo: PhotoAccessPolicy}[self.model]
        model_name = self.model._meta.model_name
        policy_model_name = policy_model._meta.model_name
        if request.POST.get("unset_access_policy"):
            queryset = queryset.select_related("access_policy")
            # Check permissions
            has_delete_perm = request.user.has_perm(
                f"gallery.delete_{policy_model_name}"
            )
            for obj in queryset:
                try:
                    obj.access_policy
                except policy_model.DoesNotExist:
                    pass
                else:
                    if not has_delete_perm:
                        raise PermissionDenied
            # Apply changes
            deleted = 0
            for obj in queryset:
                try:
                    ap = obj.access_policy
                except policy_model.DoesNotExist:
                    pass
                else:
                    ap.delete()
                    deleted += 1
            message = gettext("Successfully deleted %(deleted)d access policies.")
            message = message % {"deleted": deleted}
            self.message_user(request, message)
            return HttpResponseRedirect(
                reverse(f"admin:gallery_{model_name}_changelist")
            )

        context = {
            "action_checkbox_name": ACTION_CHECKBOX_NAME,
            "opts": self.model._meta,
            "photos": Photo.objects.filter(album__in=queryset),
            "queryset": queryset,
            "title": gettext("Unset access policy"),
        }
        return render(request, "admin/gallery/unset_access_policy.html", context)

    unset_access_policy.short_description = gettext_lazy("Unset access policy")


class AccessPolicyInline(admin.StackedInline):
    filter_horizontal = ("groups", "users")


class AlbumAccessPolicyInline(AccessPolicyInline):
    model = AlbumAccessPolicy


class AlbumAdmin(SetAccessPolicyMixin, admin.ModelAdmin):
    date_hierarchy = "date"
    inlines = (AlbumAccessPolicyInline,)
    list_display = (
        "display_name",
        "date",
        "category",
        "public",
        "groups",
        "users",
        "inherit",
    )
    list_filter = ("category",)
    ordering = ("-date", "-name", "-dirpath", "-category")
    readonly_fields = ("dirpath",)
    search_fields = ("name", "dirpath")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related("access_policy__users")
            .prefetch_related("access_policy__groups")
        )

    def public(self, obj):
        access_policy = obj.get_access_policy()
        if access_policy:
            return access_policy.public

    public.boolean = True

    def groups(self, obj):
        access_policy = obj.get_access_policy()
        if access_policy:
            groups = access_policy.groups.all()
            return ", ".join(str(group) for group in groups)
        else:
            return "-"

    def users(self, obj):
        access_policy = obj.get_access_policy()
        if access_policy:
            users = access_policy.users.all()
            return ", ".join(str(user) for user in users)
        else:
            return "-"

    def inherit(self, obj):
        access_policy = obj.get_access_policy()
        if access_policy:
            return access_policy.inherit

    inherit.boolean = True


admin.site.register(Album, AlbumAdmin)


class PhotoAccessPolicyInline(AccessPolicyInline):
    model = PhotoAccessPolicy


class PhotoAdmin(SetAccessPolicyMixin, admin.ModelAdmin):
    date_hierarchy = "date"
    inlines = (PhotoAccessPolicyInline,)
    list_display = ("display_name", "date", "preview", "public", "groups", "users")
    # Since date is mandatory on albums, '-album_date' avoids showing photos
    # without date first on some databases (PostgreSQL).
    ordering = ("-album__date", "-date", "-filename")
    readonly_fields = ("filename",)
    search_fields = ("album__name", "album__dirpath", "filename")

    def get_urls(self):
        return [
            path("scan/", scan_photos, name="gallery_scan_photos"),
        ] + super().get_urls()

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related("access_policy__users")
            .prefetch_related("access_policy__groups")
            .prefetch_related("album__access_policy__users")
            .prefetch_related("album__access_policy__groups")
        )

    preview_template = Template(
        """
<a href="{{ photo.get_absolute_url }}">
<img src="{% url 'gallery:photo-resized' preset='thumb' pk=photo.pk %}"
     width="128" height="128" alt="{{ photo }}" />
</a>"""
    )

    def preview(self, obj):
        return self.preview_template.render(Context({"photo": obj}))

    preview.allow_tags = True

    def public(self, obj):
        access_policy = obj.get_effective_access_policy()
        if access_policy:
            return access_policy.public

    public.boolean = True

    def groups(self, obj):
        access_policy = obj.get_effective_access_policy()
        if access_policy:
            return ", ".join(str(group) for group in access_policy.groups.all())
        else:
            return "-"

    def users(self, obj):
        access_policy = obj.get_effective_access_policy()
        if access_policy:
            return ", ".join(str(user) for user in access_policy.users.all())
        else:
            return "-"


admin.site.register(Photo, PhotoAdmin)


@permission_required("gallery.scan")
def scan_photos(request):
    if request.method == "POST":
        stdout, stderr = io.StringIO(), io.StringIO()
        management.call_command("scanphotos", stdout=stdout, stderr=stderr)
        for line in stdout.getvalue().splitlines():
            messages.info(request, line)
        for line in stderr.getvalue().splitlines():
            messages.error(request, line)
        return HttpResponseRedirect(reverse("admin:gallery_scan_photos"))
    context = {
        "app_label": "gallery",
        "title": gettext("Scan photos"),
    }
    return render(request, "admin/gallery/scan_photos.html", context)



import datetime
import io
import sys
import unittest

from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
from django.contrib.auth.models import Permission, User
from django.test import TestCase
from django.urls import reverse

from .models import Album, AlbumAccessPolicy, Photo, PhotoAccessPolicy
from .storages import get_storage


class AdminTests(TestCase):
    def setUp(self):
        today = datetime.date.today()
        self.album = Album.objects.create(category="default", dirpath="foo", date=today)
        self.album2 = Album.objects.create(
            category="default", dirpath="foo2", date=today
        )
        AlbumAccessPolicy.objects.create(album=self.album, public=True, inherit=False)
        self.photo = Photo.objects.create(album=self.album, filename="bar")
        self.photo2 = Photo.objects.create(album=self.album, filename="bar2")
        PhotoAccessPolicy.objects.create(photo=self.photo, public=True)
        self.user = User.objects.create_superuser("user", "user@gallery", "pass")
        self.client.login(username="user", password="pass")

    def test_album_changelist(self):
        self.client.get(reverse("admin:gallery_album_changelist"))

    def test_photo_changelist(self):
        self.client.get(reverse("admin:gallery_photo_changelist"))

    def test_album_change(self):
        self.client.get(reverse("admin:gallery_album_change", args=[self.album.pk]))

    def test_photo_change(self):
        self.client.get(reverse("admin:gallery_photo_change", args=[self.photo.pk]))

    def test_scan_photos(self):
        self.client.get(reverse("admin:gallery_scan_photos"))
        get_storage("photo").save("test", io.BytesIO(b"test"))
        self.client.post(reverse("admin:gallery_scan_photos"))

    def test_set_album_access_policy(self):
        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk), str(self.album2.pk)],
            },
        )
        self.assertTemplateUsed(response, "admin/gallery/set_access_policy.html")
        self.assertFalse(Album.objects.get(pk=self.album.pk).access_policy.inherit)
        with self.assertRaises(AlbumAccessPolicy.DoesNotExist):
            Album.objects.get(pk=self.album2.pk).access_policy

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk), str(self.album2.pk)],
                "public": True,
                "inherit": True,
                "set_access_policy": "Set access policy",
            },
        )
        self.assertRedirects(response, reverse("admin:gallery_album_changelist"))
        self.assertTrue(Album.objects.get(pk=self.album.pk).access_policy.inherit)
        self.assertTrue(Album.objects.get(pk=self.album2.pk).access_policy.inherit)

    def test_set_album_access_policy_invalid_form(self):
        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk), str(self.album2.pk)],
                "set_access_policy": "Set access policy",
                "users": ["-1"],
            },
        )
        self.assertTemplateUsed(response, "admin/gallery/set_access_policy.html")
        self.assertFormError(
            response,
            "form",
            "users",
            "Select a valid choice. -1 is not one of the available choices.",
        )

    def test_set_album_access_policy_no_add_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album2.pk)],
                "public": True,
                "inherit": True,
                "set_access_policy": "Set access policy",
            },
        )
        self.assertEqual(response.status_code, 403)
        with self.assertRaises(AlbumAccessPolicy.DoesNotExist):
            Album.objects.get(pk=self.album2.pk).access_policy

    def test_set_album_access_policy_add_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))
        self.user.user_permissions.add(
            Permission.objects.get(codename="add_albumaccesspolicy")
        )

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album2.pk)],
                "public": True,
                "inherit": True,
                "set_access_policy": "Set access policy",
            },
        )
        self.assertRedirects(response, reverse("admin:gallery_album_changelist"))
        self.assertTrue(Album.objects.get(pk=self.album2.pk).access_policy.inherit)

    def test_set_album_access_policy_no_change_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk)],
                "public": True,
                "inherit": True,
                "set_access_policy": "Set access policy",
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertFalse(Album.objects.get(pk=self.album.pk).access_policy.inherit)

    def test_set_album_access_policy_change_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))
        self.user.user_permissions.add(
            Permission.objects.get(codename="change_albumaccesspolicy")
        )

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "set_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk)],
                "public": True,
                "inherit": True,
                "set_access_policy": "Set access policy",
            },
        )
        self.assertRedirects(response, reverse("admin:gallery_album_changelist"))
        self.assertTrue(Album.objects.get(pk=self.album.pk).access_policy.inherit)

    def test_unset_album_access_policy(self):
        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "unset_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk), str(self.album2.pk)],
            },
        )
        self.assertTemplateUsed(response, "admin/gallery/unset_access_policy.html")
        self.assertTrue(Album.objects.get(pk=self.album.pk).access_policy.public)

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "unset_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk), str(self.album2.pk)],
                "public": True,
                "unset_access_policy": "Unset access policy",
            },
        )
        self.assertRedirects(response, reverse("admin:gallery_album_changelist"))
        with self.assertRaises(AlbumAccessPolicy.DoesNotExist):
            Album.objects.get(pk=self.album.pk).access_policy

    def test_unset_album_access_policy_no_delete_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "unset_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk)],
                "public": True,
                "unset_access_policy": "Unset access policy",
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertTrue(Album.objects.get(pk=self.album.pk).access_policy.public)

    def test_unset_album_access_policy_delete_permission(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(Permission.objects.get(codename="change_album"))
        self.user.user_permissions.add(
            Permission.objects.get(codename="delete_albumaccesspolicy")
        )

        response = self.client.post(
            reverse("admin:gallery_album_changelist"),
            {
                "action": "unset_access_policy",
                ACTION_CHECKBOX_NAME: [str(self.album.pk)],
                "public": True,
                "unset_access_policy": "Unset access policy",
            },
        )
        self.assertRedirects(response, reverse("admin:gallery_album_changelist"))
        with self.assertRaises(AlbumAccessPolicy.DoesNotExist):
            Album.objects.get(pk=self.album.pk).access_policy

    # See https://code.djangoproject.com/ticket/24258
    if (3, 0) <= sys.version_info[:2] < (3, 3):  # pragma: no cover
        test_set_album_access_policy = unittest.expectedFailure(
            test_set_album_access_policy
        )
        test_set_album_access_policy_add_permission = unittest.expectedFailure(
            test_set_album_access_policy_add_permission
        )  # noqa
        test_set_album_access_policy_change_permission = unittest.expectedFailure(
            test_set_album_access_policy_change_permission
        )  # noqa
        test_unset_album_access_policy = unittest.expectedFailure(
            test_unset_album_access_policy
        )
        test_unset_album_access_policy_delete_permission = unittest.expectedFailure(
            test_unset_album_access_policy_delete_permission
        )  # noqa



import os

from django.conf import settings
from django.contrib.auth.models import Group, User
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from .resizers import get_resize


class AccessPolicy(models.Model):
    public = models.BooleanField(verbose_name="is public", default=False)
    groups = models.ManyToManyField(Group, blank=True, verbose_name="authorized groups")
    users = models.ManyToManyField(User, blank=True, verbose_name="authorized users")

    class Meta:
        abstract = True

    def allows(self, user):
        if self.public:
            return True
        if user.is_authenticated:
            if set(self.groups.all()) & set(user.groups.all()):
                return True
            if user in self.users.all():
                return True
        return False


class AlbumManager(models.Manager):
    def allowed_for_user(self, user, include_public=True):
        album_cond = Q()
        if include_public:
            album_cond |= Q(access_policy__public=True)
        if user.is_authenticated:
            album_cond |= Q(access_policy__users=user)
            album_cond |= Q(access_policy__groups__user=user)
        return self.filter(album_cond).distinct()


class Album(models.Model):
    category = models.CharField(max_length=100)
    dirpath = models.CharField(max_length=200, verbose_name="directory path")
    date = models.DateField()
    name = models.CharField(max_length=100, blank=True)

    objects = AlbumManager()

    class Meta:
        ordering = ("date", "name", "dirpath", "category")
        unique_together = ("dirpath", "category")
        verbose_name = _("album")
        verbose_name_plural = _("albums")

    def __str__(self):
        return self.dirpath

    def get_absolute_url(self):
        return reverse("gallery:album", args=[self.pk])

    @property
    def display_name(self):
        return self.name or self.dirpath.replace("/", " > ")

    def get_access_policy(self):
        try:
            return self.access_policy
        except AlbumAccessPolicy.DoesNotExist:
            pass

    def is_allowed_for_user(self, user):
        access_policy = self.get_access_policy()
        return access_policy is not None and access_policy.allows(user)

    def get_next_in_queryset(self, albums):
        albums = albums.filter(
            Q(date__gt=self.date)
            | Q(date=self.date, name__gt=self.name)
            | Q(date=self.date, name=self.name, dirpath__gt=self.dirpath)
            | Q(
                date=self.date,
                name=self.name,
                dirpath=self.dirpath,
                category__gt=self.category,
            )
        )
        return albums.order_by("date", "name", "dirpath", "category")[:1].get()

    def get_previous_in_queryset(self, albums):
        albums = albums.filter(
            Q(date__lt=self.date)
            | Q(date=self.date, name__lt=self.name)
            | Q(date=self.date, name=self.name, dirpath__lt=self.dirpath)
            | Q(
                date=self.date,
                name=self.name,
                dirpath=self.dirpath,
                category__lt=self.category,
            )
        )
        return albums.order_by("-date", "-name", "-dirpath", "-category")[:1].get()


class AlbumAccessPolicy(AccessPolicy):
    album = models.OneToOneField(
        Album, on_delete=models.CASCADE, related_name="access_policy"
    )
    inherit = models.BooleanField(
        blank=True, default=True, verbose_name="photos inherit album access policy"
    )

    class Meta:
        verbose_name = _("album access policy")
        verbose_name_plural = _("album access policies")

    def __str__(self):
        return f"Access policy for {self.album}"


class PhotoManager(models.Manager):
    def allowed_for_user(self, user):
        inherit = Q(access_policy__isnull=True, album__access_policy__inherit=True)
        photo_cond = Q(access_policy__public=True)
        album_cond = Q(album__access_policy__public=True)
        if user.is_authenticated:
            photo_cond |= Q(access_policy__users=user)
            photo_cond |= Q(access_policy__groups__user=user)
            album_cond |= Q(album__access_policy__users=user)
            album_cond |= Q(album__access_policy__groups__user=user)
        return self.filter(photo_cond | (inherit & album_cond)).distinct()


class Photo(models.Model):
    album = models.ForeignKey(Album, on_delete=models.CASCADE)
    filename = models.CharField(max_length=100, verbose_name="file name")
    date = models.DateTimeField(null=True, blank=True)

    objects = PhotoManager()

    class Meta:
        ordering = ("date", "filename")
        permissions = (
            ("view", "Can see all photos"),
            ("scan", "Can scan the photos directory"),
        )
        unique_together = ("album", "filename")
        verbose_name = _("photo")
        verbose_name_plural = _("photos")

    def __str__(self):
        return self.filename

    def get_absolute_url(self):
        return reverse("gallery:photo", args=[self.pk])

    @property
    def display_name(self):
        return self.date or os.path.splitext(self.filename)[0]

    def get_effective_access_policy(self):
        try:
            return self.access_policy
        except PhotoAccessPolicy.DoesNotExist:
            pass
        try:
            album_access_policy = self.album.access_policy
        except AlbumAccessPolicy.DoesNotExist:
            pass
        else:
            if album_access_policy.inherit:
                return album_access_policy

    def is_allowed_for_user(self, user):
        access_policy = self.get_effective_access_policy()
        return access_policy is not None and access_policy.allows(user)

    # In the next two functions, images whose date is None may come
    # first or last, depending on the database.
    # These expressions are optimized for clarity, not concision.

    def get_next_in_queryset(self, photos):
        if self.date is None:
            photos = photos.filter(
                Q(date__isnull=False) | Q(date__isnull=True, filename__gt=self.filename)
            )
        else:
            photos = photos.filter(
                Q(date__gt=self.date) | Q(date=self.date, filename__gt=self.filename)
            )
        return photos.order_by("date", "filename")[:1].get()

    def get_previous_in_queryset(self, photos):
        if self.date is None:
            photos = photos.filter(date__isnull=True, filename__lt=self.filename)
        else:
            photos = photos.filter(
                Q(date__isnull=True)
                | Q(date__lt=self.date)
                | Q(date=self.date, filename__gt=self.filename)
            )
        return photos.order_by("-date", "-filename")[:1].get()

    @property
    def image_name(self):
        return os.path.join(self.album.dirpath, self.filename)

    def resized_url(self, preset):
        resize = get_resize()
        width, height, crop = settings.GALLERY_RESIZE_PRESETS[preset]
        return resize(self, width, height, crop)


class PhotoAccessPolicy(AccessPolicy):
    photo = models.OneToOneField(
        Photo, on_delete=models.CASCADE, related_name="access_policy"
    )

    class Meta:
        verbose_name = _("photo access policy")
        verbose_name_plural = _("photo access policies")

    def __str__(self):
        return f"Access policy for {self.photo}"



import datetime

from django.contrib.auth.models import Group, User
from django.test import TestCase

from .models import Album, AlbumAccessPolicy, Photo, PhotoAccessPolicy


class AccessPolicyTests(TestCase):
    def setUp(self):
        today = datetime.date.today()
        self.album = Album.objects.create(category="default", dirpath="foo", date=today)
        self.photo = Photo.objects.create(album=self.album, filename="bar")
        self.group = Group.objects.create(name="group")
        self.user = User.objects.create_user("user", "user@gallery", "pass")
        self.user.groups.add(self.group)
        self.other = User.objects.create_user("other", "other@gallery", "word")

    def assertAlbumAllowedFor(self, user):
        self.assertTrue(self.album.is_allowed_for_user(user))
        self.assertQuerysetEqual(
            Album.objects.allowed_for_user(user),
            [self.album.id],
            lambda album: album.id,
        )

    def assertAlbumNotAllowedFor(self, user):
        self.assertFalse(self.album.is_allowed_for_user(user))
        self.assertQuerysetEqual(
            Album.objects.allowed_for_user(user),
            [],
            lambda album: album.id,
        )

    def assertPhotoAllowedFor(self, user):
        self.assertTrue(self.photo.is_allowed_for_user(user))
        self.assertQuerysetEqual(
            Photo.objects.allowed_for_user(user),
            [self.photo.id],
            lambda photo: photo.id,
        )

    def assertPhotoNotAllowedFor(self, user):
        self.assertFalse(self.photo.is_allowed_for_user(user))
        self.assertQuerysetEqual(
            Photo.objects.allowed_for_user(user),
            [],
            lambda photo: photo.id,
        )

    def test_private_album(self):
        self.assertAlbumNotAllowedFor(self.user)

    def test_public_album(self):
        AlbumAccessPolicy.objects.create(album=self.album, public=True)
        self.assertAlbumAllowedFor(self.user)

    def test_user_album(self):
        policy = AlbumAccessPolicy.objects.create(album=self.album, public=False)
        policy.users.add(self.user)
        self.assertAlbumAllowedFor(self.user)
        self.assertAlbumNotAllowedFor(self.other)

    def test_group_album(self):
        policy = AlbumAccessPolicy.objects.create(album=self.album, public=False)
        policy.groups.add(self.group)
        self.assertAlbumAllowedFor(self.user)
        self.assertAlbumNotAllowedFor(self.other)

    def test_private_photo(self):
        self.assertPhotoNotAllowedFor(self.user)

    def test_public_photo(self):
        PhotoAccessPolicy.objects.create(photo=self.photo, public=True)
        self.assertPhotoAllowedFor(self.user)

    def test_user_photo(self):
        policy = PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        policy.users.add(self.user)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_group_photo(self):
        policy = PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        policy.groups.add(self.group)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_public_photo_inherit(self):
        AlbumAccessPolicy.objects.create(album=self.album, public=True)
        self.assertPhotoAllowedFor(self.user)

    def test_user_photo_inherit(self):
        policy = AlbumAccessPolicy.objects.create(album=self.album, public=False)
        policy.users.add(self.user)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_group_photo_inherit(self):
        policy = AlbumAccessPolicy.objects.create(album=self.album, public=False)
        policy.groups.add(self.group)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_public_photo_no_inherit(self):
        AlbumAccessPolicy.objects.create(album=self.album, inherit=False, public=True)
        self.assertPhotoNotAllowedFor(self.user)
        PhotoAccessPolicy.objects.create(photo=self.photo, public=True)
        self.assertPhotoAllowedFor(self.user)

    def test_user_photo_no_inherit(self):
        policy = AlbumAccessPolicy.objects.create(
            album=self.album, inherit=False, public=False
        )
        policy.users.add(self.user)
        self.assertPhotoNotAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)
        policy = PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        policy.users.add(self.user)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_group_photo_no_inherit(self):
        policy = AlbumAccessPolicy.objects.create(
            album=self.album, inherit=False, public=False
        )
        policy.groups.add(self.group)
        self.assertPhotoNotAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)
        policy = PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        policy.groups.add(self.group)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_public_photo_counter_inherit(self):
        AlbumAccessPolicy.objects.create(album=self.album, inherit=True, public=True)
        self.assertPhotoAllowedFor(self.user)
        PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        self.assertPhotoNotAllowedFor(self.user)

    def test_user_photo_counter_inherit(self):
        policy = AlbumAccessPolicy.objects.create(
            album=self.album, inherit=True, public=False
        )
        policy.users.add(self.user)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)
        PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        self.assertPhotoNotAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)

    def test_group_photo_counter_inherit(self):
        policy = AlbumAccessPolicy.objects.create(
            album=self.album, inherit=True, public=False
        )
        policy.groups.add(self.group)
        self.assertPhotoAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)
        PhotoAccessPolicy.objects.create(photo=self.photo, public=False)
        self.assertPhotoNotAllowedFor(self.user)
        self.assertPhotoNotAllowedFor(self.other)



# Resize images with Thumbor

# The implementation assumes that Thumbor is configured to access photos
# directly. See https://github.com/aaugustin/myks-thumbor for an example.

# It requires THUMBOR_SERVER and THUMBOR_SECURITY_KEY settings.

import urllib.parse

import libthumbor
from django.conf import settings


def resize(photo, width, height, crop=True):
    image_url = urllib.parse.quote(photo.image_name.encode())
    url = libthumbor.CryptoURL(key=settings.THUMBOR_SECURITY_KEY).generate(
        image_url=image_url,
        width=width,
        height=height,
        fit_in=not crop,
        smart=True,
    )
    return settings.THUMBOR_SERVER + urllib.parse.unquote(url)



import datetime
import zipfile

from django.contrib.auth.models import Permission, User
from django.test import TestCase
from django.urls import reverse

from .models import Album, AlbumAccessPolicy, Photo
from .resizers.pillow import get_resized_name
from .resizers.test_pillow import make_image
from .storages import get_storage


class ViewsTestsMixin:
    def make_image(self):
        make_image(self.photo.image_name, 48, 36, get_storage("photo"))

    def test_index_view(self):
        response = self.client.get(reverse("gallery:index"))
        self.assertTemplateUsed(response, "gallery/album_archive.html")

    def test_search_view(self):
        response = self.client.get(reverse("gallery:index"), {"q": "original"})
        self.assertTemplateUsed(response, "gallery/album_archive.html")

    def test_year_view(self):
        response = self.client.get(reverse("gallery:year", args=[self.album.date.year]))
        self.assertTemplateUsed(response, "gallery/album_archive_year.html")

    def test_album_view(self):
        response = self.client.get(reverse("gallery:album", args=[self.album.pk]))
        self.assertTemplateUsed(response, "gallery/album_detail.html")

    def test_photo_view(self):
        response = self.client.get(reverse("gallery:photo", args=[self.photo.pk]))
        self.assertTemplateUsed(response, "gallery/photo_detail.html")

    def test_album_export_view(self):
        self.make_image()
        url = reverse("gallery:album-export", args=[self.album.pk])
        response = self.client.get(url)
        self.assertTrue(response["Location"].startswith("/url/of/export/"))
        export_file = response["Location"][len("/url/of/") :]
        with zipfile.ZipFile(get_storage("cache").open(export_file)) as archive:
            self.assertEqual(archive.namelist(), ["original.jpg"])

    def test_photo_resized_view(self):
        with self.settings(GALLERY_RESIZE_PRESETS={"resized": (120, 120, False)}):
            self.make_image()
            url = reverse("gallery:photo-resized", args=["resized", self.photo.pk])
            response = self.client.get(url)
            resized_name = get_resized_name(self.photo, 120, 120, False)
            self.assertRedirects(
                response, "/url/of/" + resized_name, fetch_redirect_response=False
            )

    def test_photo_original_view(self):
        self.make_image()
        url = reverse("gallery:photo-original", args=[self.photo.pk])
        response = self.client.get(url)
        self.assertRedirects(
            response, "/url/of/" + self.photo.image_name, fetch_redirect_response=False
        )

    def test_latest_view(self):
        response = self.client.get(reverse("gallery:latest"))
        self.assertRedirects(response, reverse("gallery:album", args=[self.album.pk]))
        self.album.delete()
        response = self.client.get(reverse("gallery:latest"))
        self.assertRedirects(response, reverse("gallery:index"))


class ViewsWithPermissionTests(ViewsTestsMixin, TestCase):
    def setUp(self):
        super().setUp()
        today = datetime.date.today()
        self.album = Album.objects.create(
            category="default", dirpath="album", date=today
        )
        self.photo = Photo.objects.create(album=self.album, filename="original.jpg")
        self.user = User.objects.create_user("user", "user@gallery", "pass")
        self.user.user_permissions.add(Permission.objects.get(codename="view"))
        self.client.login(username="user", password="pass")


class ViewsWithPrivateAccessPolicyTests(ViewsTestsMixin, TestCase):
    def setUp(self):
        super().setUp()
        today = datetime.date.today()
        self.album = Album.objects.create(
            category="default", dirpath="album", date=today
        )
        AlbumAccessPolicy.objects.create(album=self.album, public=False, inherit=True)
        self.photo = Photo.objects.create(album=self.album, filename="original.jpg")
        self.user = User.objects.create_user("user", "user@gallery", "pass")
        self.album.access_policy.users.add(self.user)
        self.client.login(username="user", password="pass")

    def test_hide_private_albums(self):
        self.client.logout()
        response = self.client.get(reverse("gallery:index"))
        self.assertQuerysetEqual(response.context["latest"], [])


class ViewsWithPublicAccessPolicyTests(ViewsTestsMixin, TestCase):
    def setUp(self):
        super().setUp()
        today = datetime.date.today()
        self.album = Album.objects.create(
            category="default", dirpath="album", date=today
        )
        AlbumAccessPolicy.objects.create(album=self.album, public=True, inherit=True),
        self.photo = Photo.objects.create(album=self.album, filename="original.jpg")

    def test_show_and_hide_public_albums(self):
        # Public albums are shown for anonymous users
        response = self.client.get(reverse("gallery:index"))
        self.assertQuerysetEqual(
            response.context["latest"], [self.album], transform=lambda a: a
        )

        # By default public albums are hidden for authenticated users
        self.user = User.objects.create_user("user", "user@gallery", "pass")
        self.client.login(username="user", password="pass")
        response = self.client.get(reverse("gallery:index"))
        self.assertQuerysetEqual(response.context["latest"], [])

        # Showing public albums is persistent
        response = self.client.get(reverse("gallery:index") + "?show_public")
        self.assertQuerysetEqual(
            response.context["latest"], [self.album], transform=lambda a: a
        )
        response = self.client.get(reverse("gallery:index"))
        self.assertQuerysetEqual(
            response.context["latest"], [self.album], transform=lambda a: a
        )

        # Hiding public albums is persistent
        response = self.client.get(reverse("gallery:index") + "?hide_public")
        self.assertQuerysetEqual(response.context["latest"], [])
        response = self.client.get(reverse("gallery:index"))
        self.assertQuerysetEqual(response.context["latest"], [])



import functools
import re

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.module_loading import import_string


@functools.lru_cache()
def get_storage(name):
    storage_setting = f"GALLERY_{name.upper()}_STORAGE"
    try:
        storage = getattr(settings, storage_setting)
    except AttributeError:
        raise ImproperlyConfigured(f"Please define {storage_setting}")
    if isinstance(storage, str):
        return import_string(storage)()
    else:
        # To make testing ealier, the setting can be set to the storage itself.
        return storage


@receiver(setting_changed)
def clear_get_storage_cache(**kwargs):
    if re.match(r"^GALLERY_[A-Z]+_STORAGE$", kwargs["setting"]):
        get_storage.cache_clear()



from django.test import TestCase

from ..models import Album, Photo
from .thumbor import resize


class ResizeTests(TestCase):
    def setUp(self):
        super().setUp()
        self.album = Album(category="default", dirpath="album")
        self.photo = Photo(album=self.album, filename="original.jpg")

    def test_resize(self):
        self.assertEqual(
            resize(self.photo, 128, 128, True),
            "http://localhost:8888/UqYbfk-dJ8uNobtFuKVzxIwKCsA=/128x128/smart/album/original.jpg",  # noqa
        )

    def test_resize_no_crop(self):
        self.assertEqual(
            resize(self.photo, 128, 128, False),
            "http://localhost:8888/VKEUnVMAgbzCbjfBfwvEEM3UzGI=/fit-in/128x128/smart/album/original.jpg",  # noqa
        )



import datetime
import io

from django.test import TestCase
from PIL import Image, ImageDraw

from ..models import Album, Photo
from ..storages import get_storage
from ..test_storages import MemoryStorage
from .pillow import make_thumbnail, resize


def make_image(name, width, height, storage, *, format="JPEG", mode="RGB"):
    """
    Utility function to create an image for testing.

    """
    im = Image.new(mode, (width, height))
    draw = ImageDraw.Draw(im)
    draw.rectangle([0, 0, width // 2, height // 2], "#F00")
    draw.rectangle([width // 2, 0, width, height // 2], "#0F0")
    draw.rectangle([0, height // 2, width // 2, height], "#00F")
    draw.rectangle([width // 2, height // 2, width, height], "#000")
    draw.rectangle([width // 4, height // 4, 3 * width // 4, 3 * height // 4], "#FFF")
    im_bytes_io = io.BytesIO()
    im.save(im_bytes_io, format)
    im_bytes_io.seek(0)
    storage.save(name, im_bytes_io)


class ResizeTests(TestCase):
    def setUp(self):
        super().setUp()
        date = datetime.date(2023, 1, 1)
        self.album = Album(category="default", dirpath="album", date=date)
        self.photo = Photo(album=self.album, filename="original.jpg")
        make_image(self.photo.image_name, 48, 36, get_storage("photo"))

    def test_resize(self):
        thumb_name = "2301/9b35d14e151f31cf363cd9c1c1342b1f.jpg"
        self.assertEqual(resize(self.photo, 16, 16, True), f"/url/of/{thumb_name}")

        im = Image.open(get_storage("cache").open(thumb_name))
        self.assertEqual(im.size, (16, 16))


class ThumbnailTests(TestCase):
    def setUp(self):
        super().setUp()
        self.storage = MemoryStorage()

    def make_image(
        self, width, height, image_name="original.jpg", format="JPEG", mode="RGB"
    ):
        make_image(image_name, width, height, self.storage, format=format, mode=mode)

    def make_thumbnail(
        self, width, height, crop, image_name="original.jpg", thumb_name="thumbnail.jpg"
    ):
        make_thumbnail(
            image_name, thumb_name, width, height, crop, self.storage, self.storage
        )

    def open_image(self, thumb_name="thumbnail.jpg"):
        return Image.open(self.storage.open(thumb_name))

    def test_horizontal_thumbnail(self):
        self.make_image(48, 36)
        self.make_thumbnail(8, 8, True)

        im = self.open_image()
        self.assertEqual(im.size, (8, 8))

    def test_horizontal_preview(self):
        self.make_image(48, 36)
        self.make_thumbnail(16, 16, False)

        im = self.open_image()
        self.assertEqual(im.size, (16, 12))

    def test_square_thumbnail(self):
        self.make_image(36, 36)
        self.make_thumbnail(8, 8, True)

        im = self.open_image()
        self.assertEqual(im.size, (8, 8))

    def test_square_preview(self):
        self.make_image(36, 36)
        self.make_thumbnail(16, 16, False)

        im = self.open_image()
        self.assertEqual(im.size, (16, 16))

    def test_vertical_thumbnail(self):
        self.make_image(36, 48)
        self.make_thumbnail(8, 8, True)

        im = self.open_image()
        self.assertEqual(im.size, (8, 8))

    def test_vertical_preview(self):
        self.make_image(36, 48)
        self.make_thumbnail(16, 16, False)

        im = self.open_image()
        self.assertEqual(im.size, (12, 16))

    def test_non_jpg_thumbnail(self):
        self.make_image(36, 36, "original.png", "PNG")
        self.make_thumbnail(8, 8, True, "original.png", "thumbnail.png")

        im = self.open_image("thumbnail.png")
        self.assertEqual(im.format, "PNG")
        self.assertEqual(im.size, (8, 8))



import hashlib
import os
import random
import tempfile
import zipfile

from django.conf import settings
from django.contrib.auth.models import User
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.functional import cached_property
from django.views.generic import ArchiveIndexView, DetailView, YearArchiveView

from .models import Album, Photo
from .storages import get_storage


class GalleryCommonMixin:
    allow_future = True

    @cached_property
    def can_view_all(self):
        """
        Can the user view all photos, regardless of access policies?

        """
        return self.request.user.has_perm("gallery.view")

    @cached_property
    def show_public(self):
        """
        Should public albums be displayed, or only private ones?

        When the same gallery contains both public and private albums,
        authenticated users only see private albums by default, unless
        they have access to all photos.

        """
        session = self.request.session
        if self.request.user.is_authenticated and not self.can_view_all:
            if "show_public" in self.request.GET:
                session["show_public"] = True
            elif "hide_public" in self.request.GET:
                session["show_public"] = False
            return session.get("show_public", False)
        else:
            return True


class AlbumListMixin:
    """
    Perform access control and database optimization for albums.

    """

    model = Album
    date_field = "date"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["show_public"] = self.show_public
        return context

    def get_queryset(self):
        if self.can_view_all:
            qs = Album.objects.all()
            qs = qs.prefetch_related("photo_set")
        else:
            qs = Album.objects.allowed_for_user(self.request.user, self.show_public)
            qs = qs.prefetch_related("access_policy__groups")
            qs = qs.prefetch_related("access_policy__users")
            qs = qs.prefetch_related("photo_set__access_policy__groups")
            qs = qs.prefetch_related("photo_set__access_policy__users")
        return qs.order_by("-date", "-name")


class AlbumListWithPreviewMixin(AlbumListMixin):
    """
    Compute preview lists for albums.

    """

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user
        if not self.can_view_all and user.is_authenticated:
            # Avoid repeated queries - this is specific to django.contrib.auth
            user = User.objects.prefetch_related("groups").get(pk=user.pk)
        for album in context["object_list"]:
            if self.can_view_all:
                photos = list(album.photo_set.all())
            else:
                photos = [
                    photo
                    for photo in album.photo_set.all()
                    if photo.is_allowed_for_user(user)
                ]
            album.photos_count = len(photos)
            preview_count = getattr(settings, "GALLERY_PREVIEW_COUNT", 5)
            if len(photos) > preview_count:
                selection = sorted(
                    random.sample(range(album.photos_count), preview_count)
                )
                album.preview = [photos[index] for index in selection]
            else:
                album.preview = list(photos)
        context["title"] = getattr(settings, "GALLERY_TITLE", "Gallery")
        return context


class GalleryIndexView(GalleryCommonMixin, AlbumListWithPreviewMixin, ArchiveIndexView):
    allow_empty = True
    paginate_by = 20

    def get_queryset(self):
        qs = super().get_queryset()
        query = self.request.GET.get("q", "")
        if query:
            qs = qs.filter(name__contains=query)
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        query = self.request.GET.get("q", "")
        if query:
            context["q"] = query
        return context


class GalleryYearView(GalleryCommonMixin, AlbumListWithPreviewMixin, YearArchiveView):
    make_object_list = True
    paginate_by = 20


class AlbumView(GalleryCommonMixin, AlbumListMixin, DetailView):
    model = Album
    context_object_name = "album"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.can_view_all:
            context["photos"] = self.object.photo_set.all()
        else:
            context["photos"] = self.object.photo_set.allowed_for_user(
                self.request.user
            )
        if self.can_view_all:
            qs = Album.objects.all()
        else:
            qs = Album.objects.allowed_for_user(self.request.user, self.show_public)
        try:
            context["previous_album"] = self.object.get_previous_in_queryset(qs)
        except Album.DoesNotExist:
            pass
        try:
            context["next_album"] = self.object.get_next_in_queryset(qs)
        except Album.DoesNotExist:
            pass
        return context


class PhotoView(GalleryCommonMixin, DetailView):
    model = Photo
    context_object_name = "photo"

    def get_queryset(self):
        if self.can_view_all:
            qs = Photo.objects.all()
        else:
            qs = Photo.objects.allowed_for_user(self.request.user)
        return qs.select_related("album")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.can_view_all:
            qs = self.object.album.photo_set.all()
        else:
            qs = self.object.album.photo_set.allowed_for_user(self.request.user)
        try:
            context["previous_photo"] = self.object.get_previous_in_queryset(qs)
        except Photo.DoesNotExist:
            pass
        try:
            context["next_photo"] = self.object.get_next_in_queryset(qs)
        except Photo.DoesNotExist:
            pass
        return context


def export_album(request, pk):
    """
    Serve a zip archive containing an entire album.

    """
    album = get_object_or_404(Album, pk=pk)
    if request.user.has_perm("gallery.view"):
        photos = album.photo_set.all()
    else:
        photos = album.photo_set.allowed_for_user(request.user)

    zip_storage = get_storage("cache")
    image_storage = get_storage("photo")

    hsh = hashlib.md5()
    hsh.update(str(settings.SECRET_KEY).encode())
    hsh.update(str(pk).encode())
    for photo in photos:
        hsh.update(str(photo.pk).encode())
    zip_name = os.path.join("export", hsh.hexdigest() + ".zip")

    if not zip_storage.exists(zip_name):
        # Create the archive in a temporary file to avoid holding it in memory
        with tempfile.TemporaryFile(suffix=".zip") as temp_zip:
            with zipfile.ZipFile(temp_zip, "w") as archive:
                for photo in photos:
                    data = image_storage.open(photo.image_name).read()
                    archive.writestr(photo.filename, data)
            temp_zip.seek(0)
            zip_storage.save(zip_name, temp_zip)

    zip_url = zip_storage.url(zip_name)
    return HttpResponseRedirect(zip_url)


def _get_photo_if_allowed(request, pk):
    qs = Photo.objects
    if not request.user.has_perm("gallery.view"):
        qs = qs.allowed_for_user(request.user)
    qs = qs.select_related("album")
    return get_object_or_404(qs, pk=pk)


def resized_photo(request, preset, pk):
    """Serve a resized photo."""
    photo = _get_photo_if_allowed(request, int(pk))
    return HttpResponseRedirect(photo.resized_url(preset))


def original_photo(request, pk):
    """Serve an original photo."""
    photo = _get_photo_if_allowed(request, int(pk))
    return HttpResponseRedirect(get_storage("photo").url(photo.image_name))


def latest_album(request):
    if request.user.has_perm("gallery.view"):
        albums = Album.objects.all()
    else:
        include_public = request.session.get(
            "show_public", not request.user.is_authenticated
        )
        albums = Album.objects.allowed_for_user(request.user, include_public)
    albums = albums.order_by("-date")[:1]
    if albums:
        pk = albums[0].pk
        return HttpResponseRedirect(reverse("gallery:album", args=[pk]))
    else:
        return HttpResponseRedirect(reverse("gallery:index"))



from django.urls import path

from . import views

app_name = "gallery"

urlpatterns = [
    path("", views.GalleryIndexView.as_view(), name="index"),
    path("latest/", views.latest_album, name="latest"),
    path("year/<int:year>/", views.GalleryYearView.as_view(), name="year"),
    path("album/<int:pk>/", views.AlbumView.as_view(), name="album"),
    path("export/<int:pk>/", views.export_album, name="album-export"),
    path("photo/<int:pk>/", views.PhotoView.as_view(), name="photo"),
    path("original/<int:pk>/", views.original_photo, name="photo-original"),
    path("resized/<slug:preset>/<int:pk>/", views.resized_photo, name="photo-resized"),
]



import io
import urllib.parse

from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import Storage
from django.test import TestCase

from .storages import get_storage


class MemoryStorage(Storage):  # pragma: no cover
    """
    Limited implementation of an in-memory file storage.

    This class does the bare minimum for tests to pass. It doesn't implement
    accessed/created/modified_time.

    It doesn't provide a ``path()`` method. Its ``url()`` method returns
    ``/url/of`` followed by the file name.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.files = {}

    def _open(self, name, mode="rb"):
        assert mode == "rb"
        return io.BytesIO(self.files[name])

    def _save(self, name, content):
        content = content.read()
        assert isinstance(content, bytes)
        self.files[name] = content
        return name

    def delete(self, name):
        self.files.pop(name, None)

    def exists(self, name):
        return name in self.files

    def listdir(self, name):
        dirs, files = [], []
        for filename in sorted(self.files):
            filename = filename[len(name) :]
            if "/" in filename:
                dirs.append(filename.partition("/")[0])
            else:
                files.append(filename)
        return dirs, files

    def size(self, name):
        return len(self.files[name])

    def url(self, name):
        return "/url/of/" + urllib.parse.quote(name)


class StoragesTest(TestCase):
    def test_get_storage(self):
        with self.settings(GALLERY_FOO_STORAGE="gallery.test_storages.MemoryStorage"):
            foo_storage = get_storage("foo")
        self.assertIsInstance(foo_storage, MemoryStorage)

    def test_get_storage_caching(self):
        with self.settings(GALLERY_FOO_STORAGE="gallery.test_storages.MemoryStorage"):
            foo_storage_1 = get_storage("foo")
            foo_storage_2 = get_storage("foo")
        self.assertIs(foo_storage_1, foo_storage_2)

    def test_get_storage_unconfigured(self):
        with self.assertRaises(ImproperlyConfigured):
            get_storage("foo")



import hashlib
import io
import os.path

from django.conf import settings
from PIL import Image, ImageFile

from ..storages import get_storage


def resize(photo, width, height, crop=True):
    image_name = photo.image_name
    resized_name = get_resized_name(photo, width, height, crop)
    photo_storage = get_storage("photo")
    cache_storage = get_storage("cache")
    if not cache_storage.exists(resized_name):
        make_thumbnail(
            image_name, resized_name, width, height, crop, photo_storage, cache_storage
        )
    return cache_storage.url(resized_name)


def get_resized_name(photo, width, height, crop):
    prefix = photo.album.date.strftime("%y%m")
    hsh = hashlib.md5()
    hsh.update(str(settings.SECRET_KEY).encode())
    hsh.update(str(photo.album.pk).encode())
    hsh.update(str(photo.pk).encode())
    hsh.update(str((width, height, crop)).encode())
    ext = os.path.splitext(photo.filename)[1].lower()
    return os.path.join(prefix, hsh.hexdigest() + ext)


exif_rotations = [  # pragma: no cover
    None,
    lambda image: image,
    lambda image: image.transpose(Image.FLIP_LEFT_RIGHT),
    lambda image: image.transpose(Image.ROTATE_180),
    # shortcut for image.transpose(Image.ROTATE_180).transpose(FLIP_LEFT_RIGHT)
    lambda image: image.transpose(Image.FLIP_TOP_BOTTOM),
    lambda image: image.transpose(Image.ROTATE_270).transpose(
        Image.FLIP_LEFT_RIGHT
    ),  # noqa
    lambda image: image.transpose(Image.ROTATE_270),
    lambda image: image.transpose(Image.ROTATE_90).transpose(
        Image.FLIP_LEFT_RIGHT
    ),  # noqa
    lambda image: image.transpose(Image.ROTATE_90),
]


def make_thumbnail(
    image_name,
    resized_name,
    thumb_width,
    thumb_height,
    crop,
    image_storage,
    thumb_storage,
):

    options = getattr(settings, "GALLERY_RESIZE_OPTIONS", {})

    # Load the image
    image = Image.open(image_storage.open(image_name))
    format = image.format

    if format == "JPEG":
        # Auto-rotate JPEG files based on EXIF information
        try:  # pragma: no cover
            # Use of an undocumented API — let's catch exceptions liberally
            orientation = image._getexif()[274]
            image = exif_rotations[orientation](image)
        except Exception:
            pass

        # Increase Pillow's buffer from 64k to 4MB to avoid JPEG save errors
        ImageFile.MAXBLOCK = 4194304

    # Pre-crop if requested and the aspect ratios don't match exactly
    image_width, image_height = image.size
    if crop:
        if thumb_width * image_height > image_width * thumb_height:
            target_height = image_width * thumb_height // thumb_width
            top = (image_height - target_height) // 2
            image = image.crop((0, top, image_width, top + target_height))
        elif thumb_width * image_height < image_width * thumb_height:
            target_width = image_height * thumb_width // thumb_height
            left = (image_width - target_width) // 2
            image = image.crop((left, 0, left + target_width, image_height))

    # Resize
    image.thumbnail((thumb_width, thumb_height), Image.ANTIALIAS)

    # Save the thumbnail
    thumb_bytes_io = io.BytesIO()
    image.save(thumb_bytes_io, format, **options.get(image.format, {}))
    thumb_bytes_io.seek(0)
    thumb_storage.save(resized_name, thumb_bytes_io)



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

# Django

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

INSTALLED_APPS = [
    "gallery.apps.GalleryConfig",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
]

LOGGING = {
    "version": 1,
    "handlers": {
        "null": {
            "class": "logging.NullHandler",
        },
    },
    "loggers": {
        "django.request": {
            "handlers": ["null"],
            "propagate": False,
        },
    },
}

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "gallery.test_urls"

SECRET_KEY = "Not empty for tests."

STATIC_ROOT = BASE_DIR

STATIC_URL = "/static/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "gallery" / "test_templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# libthumbor

THUMBOR_SERVER = "http://localhost:8888"

THUMBOR_SECURITY_KEY = "thumbor-security-key"

# myks-gallery

GALLERY_PHOTO_STORAGE = "gallery.test_storages.MemoryStorage"

GALLERY_CACHE_STORAGE = "gallery.test_storages.MemoryStorage"



import collections
import datetime
import os
import re
import time
import unicodedata

from django.conf import settings
from django.core.management import base
from django.db import transaction
from django.utils import timezone

from ...models import Album, Photo
from ...storages import get_storage


class Command(base.BaseCommand):
    help = "Scan photos and update database."

    def add_arguments(self, parser):
        parser.add_argument(
            "--full",
            action="store_true",
            dest="full_sync",
            default=False,
            help="Perform a full resynchronization",
        )

    @transaction.atomic
    def handle(self, **options):
        self.full_sync = options["full_sync"]
        self.verbosity = int(options["verbosity"])

        t = time.time()

        self.write_out("Scanning photos...", verbosity=1)
        albums = scan_photo_storage(self)

        self.write_out("Synchronizing albums...", verbosity=1)
        synchronize_albums(albums, self)

        self.write_out("Synchronizing photos...", verbosity=1)
        synchronize_photos(albums, self)

        dt = time.time() - t
        self.write_out(f"Done ({dt:02f}s)", verbosity=1)

    def write_err(self, message, verbosity):
        if self.verbosity >= verbosity:
            self.stderr.write(self.style.ERROR(message + "\n"))

    def write_out(self, message, verbosity):
        if self.verbosity >= verbosity:
            self.stdout.write(message + "\n")


ignores = [re.compile(i) for i in getattr(settings, "GALLERY_IGNORES", ())]


def is_ignored(path):
    return any(pat.match(path) for pat in ignores)


patterns = [
    (cat, re.compile(pat)) for cat, pat in getattr(settings, "GALLERY_PATTERNS", ())
]


def is_matched(path):
    for category, pattern in patterns:
        match = pattern.match(path)
        if match:
            return category, match.groupdict()


def walk_photo_storage(storage, path=""):
    """
    Yield (directory path, file names) 2-uples for the photo storage.

    This function directories that do not contain any files.

    """
    directories, files = storage.listdir(path)
    if files:
        yield path, files
    for directory in directories:
        dir_path = os.path.join(path, directory)
        for result in walk_photo_storage(storage, dir_path):
            yield result


def iter_photo_storage(command):
    """
    Yield (relative path, category, regex captures) for each photo.

    """
    photo_storage = get_storage("photo")
    for dirpath, filenames in walk_photo_storage(photo_storage):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            # HFS+ stores names in NFD which causes issues with some fonts.
            filepath = unicodedata.normalize("NFKC", filepath)
            if is_ignored(filepath):
                command.write_out(f"- {filepath}", verbosity=3)
                continue
            result = is_matched(filepath)
            if result is not None:
                command.write_out(f"> {filepath}", verbosity=3)
                category, captures = result
                yield filepath, category, captures
            else:
                command.write_err(f"? {filepath}", verbosity=1)


def scan_photo_storage(command):
    """
    Return a dictionary of albums keyed by (category, dirpath).

    Each album is a dictionary of photos, keyed by filename.

    The result can be passed to ``synchronize_albums`` and
    ``synchronize_photos``.

    """
    albums = collections.defaultdict(lambda: {})
    for path, category, captures in iter_photo_storage(command):
        dirpath, filename = os.path.split(path)
        albums[category, dirpath][filename] = captures
    return albums


def get_album_info(captures, command):
    """
    Return the date and name of an album.

    ``captures`` are elements extracted from the file name of a random photo
    in the album.

    """
    date = None
    try:
        kwargs = {k: int(captures["a_" + k]) for k in ("year", "month", "day")}
        date = datetime.date(**kwargs)
    except KeyError:
        pass
    except ValueError as e:
        command.write_err(f"{e} {kwargs}", verbosity=1)
    name = " ".join(
        v
        for k, v in sorted(captures.items())
        if k.startswith("a_name") and v is not None
    )
    name = name.replace("/", " > ")
    return date, name


def get_photo_info(captures, command):
    """
    Return the datetime of a photo.

    ``captures`` are elements extracted from the file name of the photo.

    """
    date = None
    try:
        kwargs = {
            k: int(captures["p_" + k])
            for k in ("year", "month", "day", "hour", "minute", "second")
        }
        date = datetime.datetime(**kwargs)
        if settings.USE_TZ:
            date = timezone.make_aware(date, timezone.get_default_timezone())
    except KeyError:
        pass
    except ValueError as e:
        command.write_err(f"{e} {kwargs}", verbosity=1)
    return date


def synchronize_albums(albums, command):
    """
    Synchronize albums from the filesystem to the database.

    ``albums`` is the result of ``scan_photo_storage``.

    """
    new_keys = set(albums.keys())
    old_keys = set((a.category, a.dirpath) for a in Album.objects.all())
    for category, dirpath in sorted(new_keys - old_keys):
        random_capture = next(iter(albums[category, dirpath].values()))
        date, name = get_album_info(random_capture, command)
        command.write_out(f"Adding album {dirpath} ({category}) as {name}", verbosity=1)
        Album.objects.create(category=category, dirpath=dirpath, date=date, name=name)
    for category, dirpath in sorted(old_keys - new_keys):
        command.write_out(f"Removing album {dirpath} ({category})", verbosity=1)
        Album.objects.get(category=category, dirpath=dirpath).delete()


def synchronize_photos(albums, command):
    """
    Synchronize photos from the filesystem to the database.

    ``albums`` is the result of ``scan_photo_storage``.

    """
    for (category, dirpath), filenames in albums.items():
        album = Album.objects.get(category=category, dirpath=dirpath)
        new_keys = set(filenames.keys())
        old_keys = set(p.filename for p in album.photo_set.all())
        for filename in sorted(new_keys - old_keys):
            date = get_photo_info(albums[category, dirpath][filename], command)
            command.write_out(
                f"Adding photo {filename} to album {dirpath} ({category})", verbosity=2
            )
            photo = Photo.objects.create(album=album, filename=filename, date=date)
        for filename in sorted(old_keys - new_keys):
            command.write_out(
                f"Removing photo {filename} from album {dirpath} ({category})",
                verbosity=2,
            )
            photo = Photo.objects.get(album=album, filename=filename)
            photo.delete()
        if not command.full_sync:
            continue
        for filename in sorted(old_keys & new_keys):
            date = get_photo_info(albums[category, dirpath][filename], command)
            photo = Photo.objects.get(album=album, filename=filename)
            if date != photo.date:
                command.write_out(
                    f"Fixing date of photo {filename} "
                    f"from album {dirpath} ({category})",
                    verbosity=2,
                )
                photo.date = date
                photo.save()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gallery',
 'gallery.management',
 'gallery.management.commands',
 'gallery.migrations',
 'gallery.resizers']

package_data = \
{'': ['*'],
 'gallery': ['locale/fr/LC_MESSAGES/*',
             'static/css/*',
             'templates/admin/gallery/*',
             'templates/gallery/*',
             'test_templates/*']}

install_requires = \
['django>=3.2']

setup_kwargs = {
    'name': 'myks-gallery',
    'version': '0.9',
    'description': 'Photo gallery with granular access control',
    'long_description': 'mYk\'s gallery\n#############\n\nIntroduction\n============\n\n`myks-gallery`_ is a simple photo gallery with granular access control.\n\nIt powers my `humble photo gallery`_, allowing me to:\n\n- access my entire photo collection privately,\n- share some albums with family or friends,\n- make some albums public.\n\n.. _myks-gallery: https://github.com/aaugustin/myks-gallery\n.. _humble photo gallery: https://myks.org/photos/\n\nUse case\n========\n\nRather than use a photo manager, I just create a new directory for each event\nand put my photos inside. I include the date of the event in the name of the\ndirectory and I rename photos based on their date and time. Then I regularly\nsynchronize my collection to a remote storage. I serve my gallery from there.\n\nIf you have a similar workflow, you may find myks-gallery useful.\n\nWhenever I upload new photos, I re-scan the collection with ``django-admin\nscanphotos`` or with the button in the admin. myks-gallery detects new albums\nand photos. Then I define users, groups and access policies in the admin.\n\nAlbum access policies control the visibility of albums. Most often, you\'ll\nenable the "photos inherit album access policy" option. If you need more\ncontrol, for instance to share only a subset of an album, you can disable this\noption and use photo access policies. You still need to define an album access\npolicy and it should be a superset of the photo access policies.\n\nObviously, requiring usernames and passwords doesn\'t work well for sharing\nphotos with relatives. You might want to use django-sesame_.\n\n.. _django-sesame: https://github.com/aaugustin/django-sesame\n\nSetup\n=====\n\nmyks-gallery is a pluggable Django application. It requires:\n\n* Django ≥ 3.2 (LTS)\n* Python ≥ 3.6\n\nArchitecture\n------------\n\nmyks-gallery requires two storage areas:\n\n- The first one contains the original photos. It\'s a read-only reference. You\n  can upload photos with `aws s3 sync`_, `gsutil rsync`_, etc.\n- The second one contains resized versions and ZIP archives of album exports.\n  It\'s a cache. You can set up expiry policies and clear it without affecting\n  the gallery, aside from the cost of rescaling images again.\n\nmyks-gallery accesses them through Django\'s `file storage API`_, meaning that\nyou can use any storage for which a Django storage backend exists. You should\nuse a third-party storage backend if you\'re storing files in a cloud service.\n\n.. _aws s3 sync: https://docs.aws.amazon.com/cli/latest/reference/s3/sync.html\n.. _gsutil rsync: https://cloud.google.com/storage/docs/gsutil/commands/rsync\n.. _file storage API: https://docs.djangoproject.com/en/stable/ref/files/storage/\n\nInstallation guide\n------------------\n\nThis application isn\'t exactly plug\'n\'play. There are many moving pieces.\nHere\'s the general process for integrating myks-gallery into an existing\nwebsite:\n\n1.  Download and install the package from PyPI::\n\n        $ pip install myks-gallery Pillow\n\n2.  Add ``gallery.apps.GalleryConfig`` to ``INSTALLED_APPS``::\n\n        INSTALLED_APPS += [\'gallery.apps.GalleryConfig\']\n\n3.  Configure the settings — see below for the list.\n\n4.  Add the application to your URLconf with the ``gallery`` application\n    namespace::\n\n        urlpatterns += [\n            path(\'gallery/\', include(\'gallery.urls\', namespace=\'gallery\')),\n        ]\n\n5.  Create a suitable ``base.html`` template. It must provide three blocks:\n    ``title``, ``extrahead``, ``content``, as shown in this `example`_.\n\n6.  Scan your photos with the "Scan photos" button in the admin or the\n    ``scanphotos`` management command and define access policies.\n\nThe source_ contains a sample application in the ``example`` directory. It can\nhelp you see how everything fits together. See below for how to run it.\n\n.. _example: https://github.com/aaugustin/myks-gallery/blob/master/example/example/templates/base.html\n.. _X-accel: http://wiki.nginx.org/X-accel\n.. _mod_xsendfile: https://tn123.org/mod_xsendfile/\n.. _source: https://github.com/aaugustin/myks-gallery\n\nAccess control\n--------------\n\nmyks-gallery provides two levels of access control: by album and by photo.\n\nBy default, albums and photos aren\'t visible by anyone, except users with the\n"Can see all photos" permission, including superusers who have it implicitly.\n\nTo make them visible, you must define an access policy. You have two options:\npublic access or access restricted to select users or groups.\n\nAccess policies for albums are configured explicitly in the admin.\n\nIn most cases, you will enable the "Photos inherit album access policy"\noption, so that the access policy also applies to all photos in the album.\n\nAccess policies for photos may also be configured for granular control.\n\nFor example, if you want to publish just a few photos in an album, make these\nphotos public, make the album public, but don\'t enable "Photos inherit album\naccess policy". Other photos in the album won\'t be visible.\n\nAnother example, if you want to share an album privately except for a few\nphotos, set an empty access policy on these photos (e.g. by adding then\nremoving yourself), then allow some groups or users to view the album.\n\nPermissions\n-----------\n\nmyks-gallery defines two permissions for ``django.contrib.auth``:\n\n- "Can scan the photos directory" allows using the "Scan photos" button in the\n  admin.\n- "Can see all photos" allows seeing all albums and all photos regardless of\n  access policies.\n\nSettings\n--------\n\n``GALLERY_PHOTO_STORAGE``\n.........................\n\nDefault: *not defined*\n\nDotted Python path to the Django storage class for the original photos. It\nmust be readable by the application server but should not be writable.\n\nWhile ``GALLERY_PHOTO_STORAGE`` behaves like Django\'s ``DEFAULT_FILE_STORAGE``\nsetting, you\'ll usullay point it to a factory function that initializes and\nreturns a Django storage instance because you won\'t want to use globally\nconfigured values for settings such as the S3 bucket name.\n\n``GALLERY_CACHE_STORAGE``\n.........................\n\nDefault: *not defined*\n\nDotted Python path to the Django storage class for resized versions and album\narchives. It must be readable and writable by the application server.\n\nIt behaves like ``GALLERY_PHOTO_STORAGE``.\n\n``GALLERY_PATTERNS``\n....................\n\nDefault: ``()``\n\nTuple of (category name, regular expression) defining how to extract album\ninformation — category, date, name — from the paths of photo files.\n\nThe regular expressions match paths relative to the root of the photo storage.\nThey contain the following captures:\n\n- ``a_name``: album name (mandatory) — to capture several bits, use\n  ``a_name1``, ``a_name2``, etc.\n- ``a_year``, ``a_month``, ``a_day``: album date (mandatory)\n- ``p_year``, ``p_month``, ``p_day``, ``p_hour``, ``p_minute``, ``p_second``:\n  photo date and time (optional)\n\nHere\'s an example, for photos stored with names such as ``2013/01_19_Snow in\nParis/2013-01-19_19-12-43.jpg``::\n\n    GALLERY_PATTERNS = (\n        (\'Photos\',\n            r\'(?P<a_year>\\d{4})/(?P<a_month>\\d{2})_(?P<a_day>\\d{2})_\'\n            r\'(?P<a_name>[^_/]+)/\'\n            r\'(?P<p_year>\\d{4})-(?P<p_month>\\d{2})-(?P<p_day>\\d{2})_\'\n            r\'(?P<p_hour>\\d{2})-(?P<p_minute>\\d{2})-(?P<p_second>\\d{2})\\.jpg\'),\n    )\n\n``GALLERY_IGNORES``\n...................\n\nDefault: ``()``\n\nTuple of regular expressions matching paths within ``GALLERY_PHOTO_STORAGE``.\nFiles matching one of these expressions will be ignored when scanning photos.\n\n``GALLERY_RESIZE``\n..................\n\nDefault: ``gallery.resizers.pillow.resize``\n\nDotted Python path to the callable providing resizing functionality.\n\n``resize(photo, width, height, crop=True)`` receives an instance of the\n``Photo`` model and returns a URL for the resized version. The original image is\nfound at ``photo.image_name`` in the photo storage.\n\nThe default implementation depends on ``Pillow``.\n\n``GALLERY_RESIZE_PRESETS``\n..........................\n\nDefault: ``{}``\n\nDictionary mapping resizing presets to ``(width, height, crop)`` tuples. If\n``crop`` is ``True``, the photo will be cropped and the thumbnail will have\nexactly the requested size. If ``crop`` is ``False``, the thumbnail will be\nsmaller than the requested size in one dimension to preserve the photo\'s\naspect ratio.\n\nThe default templates assume the following values::\n\n    GALLERY_RESIZE_PRESETS = {\n        \'thumb\': (128, 128, True),\n        \'standard\': (768, 768, False),\n    }\n\nYou may double these sizes for better results on high DPI displays.\n\nThe admin expects a ``\'thumb\'`` preset.\n\n``GALLERY_RESIZE_OPTIONS``\n..........................\n\nDefault: ``{}``\n\nDictionary mapping image formats names to dictionaries of options for Pillow\'s\n``save`` method. Pillow\'s documentation describes options for each file format.\n\nThe following a reasonable value for high-quality thumbnails and previews::\n\n    GALLERY_RESIZE_OPTIONS = {\n        \'JPEG\': {\'quality\': 90, \'optimize\': True},\n    }\n\nThe default resizer honors this setting. Other resizers may ignore it.\n\n``GALLERY_TITLE``\n.................\n\nDefault: ``"Gallery"``\n\nTitle of your photo gallery. This is only used by the default templates of the\nindex and year views.\n\n``GALLERY_PREVIEW_COUNT``\n.........................\n\nDefault: ``5``\n\nNumber of thumbnails shown in the preview of each album.\n\nRunning the sample application\n==============================\n\n1.  Make sure Django and Pillow are installed.\n\n2.  Create directories for storing photos and thumbnails::\n\n        $ cd example\n        $ mkdir media\n        $ mkdir media/cache\n        $ mkdir media/photos\n\n3.  Create an album directory, whose name must contain a date, and download\n    images. `Wikipedia\'s featured pictures`_ are a good choice::\n\n        $ mkdir "media/photos/2023_01_01_Featured Pictures"\n        # ... download pictures to this directory...\n\n    .. _Wikipedia\'s featured pictures: https://en.wikipedia.org/wiki/Wikipedia:Featured_pictures\n\n4.  Run the development server::\n\n    $ ./manage.py migrate\n    $ ./manage.py createsuperuser\n    $ ./manage.py runserver\n\n5.  Go to http://localhost:8000/admin/gallery/album/ and log in. Click the\n    "Scan photos" link at the top right, and the "Scan photos" button on the\n    next page. You should see the following messages:\n\n    * Scanning path/to/myks-gallery/example/media/photos\n    * Adding album 2023_01_01_Featured Pictures (Photos) as Featured Pictures\n    * Done (0.01s)\n\n    Go to http://localhost:8000/ and enjoy!\n\n    Since you\'re logged in as an admin user, you can view albums and photos\n    even though you haven\'t defined any access policies yet.\n\nChangelog\n=========\n\n1.0\n---\n\n*Under development*\n\n0.9\n---\n\nThis version makes it possible to customize image resizing, for example to\nintegrate an external service, with the ``GALLERY_RESIZE`` setting. Review\n``gallery.resizers.thumbor.resize`` for an example.\n\nSeveral features designed for storing files in the filesystem are removed:\n\n* The ``--resize`` option of ``django-admin scanphotos`` is removed.\n* Expiration of album archives with the ``GALLERY_ARCHIVE_EXPIRY`` setting is\n  removed. Configure lifecycle for the ``export`` folder on the cloud storage\n  service instead.\n* Fallback to the ``GALLERY_PHOTO_DIR`` and ``GALLERY_CACHE_DIR`` settings,\n  deprecated in version 0.5, is removed.\n* Support for serving files privately from the local filesystem is removed,\n  including the ``GALLERY_SENDFILE_HEADER`` and ``GALLERY_SENDFILE_PREFIX``\n  settings.\n\nIt includes smaller changes too.\n\n* Updated for Django 4.0.\n\n0.8\n---\n\n* Changed photo access policies to always override album access policies, even\n  when "Photos inherit album access policy" is enabled. This makes it possible\n  to restrict access with photo access policies, rather than just extend it.\n* Updated for Django 3.0.\n\n0.7\n---\n\n* Updated for Django 2.0.\n\n0.6\n---\n\n* Added migrations for compatibility with Django 1.9.\n\nTo upgrade an existing project, run: ``django-admin migrate --fake-initial``.\n\n0.5\n---\n\nThis version uses the Django file storage API for all operations on files,\nmaking it possible to use services such as Amazon S3 or Google Cloud Storage\nfor storing photos and thumbnails. It introduces the ``GALLERY_PHOTO_STORAGE``\nand ``GALLERY_CACHE_STORAGE`` settings. They supersede ``GALLERY_PHOTO_DIR``\nand ``GALLERY_CACHE_DIR``.\n\nWhen upgrading to 0.5 or later, you should clear the cache directory.\nPreviously cached thumbnails and exports won\'t be used by this version.\n\nIt includes smaller changes too.\n\n* Switched ordering of albums to always show the most recent albums first.\n* Allowed customizing the number of photos in album previews.\n* Preserved original order of photos in album previews.\n* Added pagination on album preview pages.\n* Changed the hashing schema. This invalides the cache. You should clear it.\n* Fixed collision between zip archives containing photos with the same name.\n\n0.4\n---\n\n* Provided exports of albums as zip archives.\n* Fixed preview of photos affected by batch access policy changes.\n\n0.3\n---\n\n* Support for Python 3 and Django 1.6.\n* Hid public albums by default for logged-in users.\n* Switched the default styles to a responsive design.\n* Added an option to scanphotos to precompute thumbnails.\n* Added an option to scanphotos to resynchronize photo dates.\n* Fixed bugs in photo dates.\n\n0.2\n---\n\n* Made most settings optional for easier deployment.\n* Made "sendfile" optional and used streaming responses as a fallback.\n* Worked around a crash in libjpeg when creating large JPEG previews.\n* Added many tests.\n\n0.1\n---\n\n* Initial public release, with the history from my private repository.\n* Added documentation (README file).\n* Added a sample application.\n',
    'author': 'Aymeric Augustin',
    'author_email': 'aymeric.augustin@m4x.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/aaugustin/myks-gallery',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import functools

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.module_loading import import_string


@functools.lru_cache()
def get_resize():
    resize = getattr(settings, "GALLERY_RESIZE", "gallery.resizers.pillow.resize")
    return import_string(resize)


@receiver(setting_changed)
def clear_get_resize_cache(**kwargs):
    if kwargs["setting"] == "GALLERY_RESIZE":
        get_resize.cache_clear()



# Generated by Django 1.11.16 on 2018-08-17 20:28
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0008_alter_user_username_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='Album',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(max_length=100)),
                ('dirpath', models.CharField(max_length=200, verbose_name='directory path')),
                ('date', models.DateField()),
                ('name', models.CharField(blank=True, max_length=100)),
            ],
            options={
                'verbose_name': 'album',
                'verbose_name_plural': 'albums',
                'ordering': ('date', 'name', 'dirpath', 'category'),
            },
        ),
        migrations.CreateModel(
            name='AlbumAccessPolicy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public', models.BooleanField(default=False, verbose_name='is public')),
                ('inherit', models.BooleanField(blank=True, default=True, verbose_name='photos inherit album access policy')),
                ('album', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='access_policy', to='gallery.Album')),
                ('groups', models.ManyToManyField(blank=True, to='auth.Group', verbose_name='authorized groups')),
                ('users', models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL, verbose_name='authorized users')),
            ],
            options={
                'verbose_name': 'album access policy',
                'verbose_name_plural': 'album access policies',
            },
        ),
        migrations.CreateModel(
            name='Photo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('filename', models.CharField(max_length=100, verbose_name='file name')),
                ('date', models.DateTimeField(blank=True, null=True)),
                ('album', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='gallery.Album')),
            ],
            options={
                'verbose_name': 'photo',
                'verbose_name_plural': 'photos',
                'ordering': ('date', 'filename'),
                'permissions': (('view', 'Can see all photos'), ('scan', 'Can scan the photos directory')),
            },
        ),
        migrations.CreateModel(
            name='PhotoAccessPolicy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public', models.BooleanField(default=False, verbose_name='is public')),
                ('groups', models.ManyToManyField(blank=True, to='auth.Group', verbose_name='authorized groups')),
                ('photo', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='access_policy', to='gallery.Photo')),
                ('users', models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL, verbose_name='authorized users')),
            ],
            options={
                'verbose_name': 'photo access policy',
                'verbose_name_plural': 'photo access policies',
            },
        ),
        migrations.AlterUniqueTogether(
            name='album',
            unique_together=set([('dirpath', 'category')]),
        ),
        migrations.AlterUniqueTogether(
            name='photo',
            unique_together=set([('album', 'filename')]),
        ),
    ]
